import os
from contextlib import asynccontextmanager

import anyio
from fastapi import FastAPI, Header, Request
from fastapi.exceptions import HTTPException
from fastapi.responses import ORJSONResponse
from starlette_compress import CompressMiddleware

from mockai import anthropic, openai
from mockai.dependencies import ResponseFile, star_watching_responses
from mockai.gui.router import gui_router


@asynccontextmanager
async def lifespan(app: FastAPI):
    mockai_embedding_size = os.getenv("MOCKAI_EMBEDDING_SIZE", "1536")
    app.state.embedding_size = int(mockai_embedding_size)

    async with anyio.create_task_group() as tg:
        tg.start_soon(star_watching_responses)
        yield


app = FastAPI(lifespan=lifespan, default_response_class=ORJSONResponse)

app.add_middleware(CompressMiddleware)

app.include_router(openai._router)
app.include_router(anthropic._router)
app.include_router(gui_router)


@app.post("/{path:path}")
async def route_response(
    path: str,
    data: dict,
    request: Request,
    responses: ResponseFile,
    mock_response: str | None = Header(default=None),
    user_agent: str | None = Header(default=None),
):
    if not user_agent:
        raise HTTPException(400, "User agent could not be determined")

    if "OpenAI" in user_agent:
        if "completions" in path:
            return await openai._generate_openai_completion_response(
                data, responses, mock_response
            )
        elif "embeddings" in path:
            return await openai._generate_openai_embeddings_response(
                request.app.state.embedding_size, data
            )
        else:
            return HTTPException(400, "Invalid path")
    elif "Anthropic" in user_agent:
        return await anthropic._generate_anthropic_response(
            data, responses, mock_response
        )
    else:
        raise HTTPException(400, "Invalid user agent")



from fastapi import APIRouter, Header, HTTPException
from fastapi.responses import JSONResponse, StreamingResponse

from mockai.dependencies import ResponseFile

from .models import Payload
from .services import generate_anthropic_response

anthropic_router = APIRouter(prefix="/anthropic")


@anthropic_router.post("/v1/messages")
async def anthropic_messages(
    payload: Payload,
    file: ResponseFile,
    mock_response: str | None = Header(default=None),
):
    try:
        response = await generate_anthropic_response(
            payload.model_dump(), file, mock_response
        )

        if payload.stream is True:
            return StreamingResponse(response)
        else:
            return JSONResponse(response)

    except ValueError as e:
        raise HTTPException(400, str(e))



import pytest
from testcontainers.core.container import DockerContainer
from testcontainers.core.image import DockerImage
from testcontainers.core.waiting_utils import wait_for_logs


@pytest.fixture(scope="session", autouse=True)
def mockai_server():
    with DockerImage(
        path=".", dockerfile_path="tests/test.Dockerfile", tag="ajac-zero/mock-ai:test"
    ) as image:
        with DockerContainer(str(image)).with_bind_ports(8100, 8100) as container:
            wait_for_logs(
                container, "Uvicorn running on http://0.0.0.0:8100", raise_on_exit=True
            )
            yield container



import logging
import os
from typing import Annotated

import aiofiles
import watchfiles
from fastapi import Depends
from pydantic_core._pydantic_core import ValidationError

from mockai.models.json_file.models import PreDeterminedResponses

logger = logging.getLogger(__name__)
responses = None


async def save_reload(path: str, change: watchfiles.Change):
    global responses

    try:
        async with aiofiles.open(path) as f:
            responses = PreDeterminedResponses.model_validate_json(await f.read())
        logger.info(
            "Predetermined responses reloaded after %s",
            "modification" if change == watchfiles.Change.modified else "recreation",
        )
    except ValidationError as e:
        error = e.errors()[0]
        logger.error(
            f"Error reloading responses\n"
            f"Problematic input: {error['input']}\n"
            f"Fix: {error['msg']}\n"
            f"Keeping old responses"
        )


async def star_watching_responses():
    global responses
    if file := os.getenv("MOCKAI_RESPONSES"):
        async with aiofiles.open(file) as f:
            contents = await f.read()
        responses = PreDeterminedResponses.model_validate_json(contents)

        try:
            async for changes in watchfiles.awatch(file):
                change = changes.pop()
                change, _ = change
                if (
                    change == watchfiles.Change.modified
                    or change == watchfiles.Change.added
                ):
                    await save_reload(file, change)
                else:
                    logger.info(
                        "Predetirmened response file deleted. Clearing responses"
                    )
                    responses = None
        except Exception:
            logger.exception("Unexpected Error watching responses file")
            raise


async def get_responses() -> PreDeterminedResponses | None:
    return responses


ResponseFile = Annotated[PreDeterminedResponses | None, Depends(get_responses)]



from fastapi import APIRouter, Header, HTTPException, Request
from fastapi.responses import JSONResponse, StreamingResponse

from mockai.dependencies import ResponseFile

from . import models, services

openai_router = APIRouter(prefix="/openai")


@openai_router.post("/chat/completions")  # OpenAI Endpoint
@openai_router.post("/deployments/{path}/chat/completions")  # AzureOpenAI Endpoint
async def openai_chat_completion(
    payload: models.Payload,
    responses: ResponseFile,
    mock_response: str | None = Header(default=None),
):
    try:
        response = await services.generate_openai_completion_response(
            payload.model_dump(), responses, mock_response
        )

        if payload.stream is True:
            return StreamingResponse(response)
        else:
            return JSONResponse(response)
    except ValueError as e:
        raise HTTPException(400, str(e))


@openai_router.post("/embeddings")  # OpenAI Endpoint
@openai_router.post("/deployments/{path}/embeddings")  # AzureOpenAI Endpoint
async def openai_create_embeddings(request: Request, payload: models.EmbeddingPayload):
    return await services.generate_openai_embeddings_response(
        request.app.state.embedding_size, payload.model_dump()
    )



import json
import os
import subprocess

import click
import pydantic

from mockai.models.json_file.models import PreDeterminedResponses

dir_path = os.path.dirname(os.path.realpath(__file__))


@click.group()
def cli():
    pass


@cli.command()
@click.argument("responses", type=click.File("rb"), required=False)
@click.option("--embedding-size", "-E", default=1536)
@click.option("--host", "-h", default="127.0.0.1")
@click.option("--port", "-p", default=8100)
def server(responses, embedding_size, host, port):
    if responses:
        print(f"Reading pre-determined responses from {responses.name}.")  # noqa: T201

        try:
            responses_data = json.load(responses)
        except json.JSONDecodeError:
            raise click.BadParameter("Error reading JSON file: Is it valid JSON?")

        try:
            PreDeterminedResponses.model_validate(responses_data)
        except pydantic.ValidationError as e:
            error = e.errors()[0]
            raise click.BadParameter(
                f"Error validating responses. Make sure they follow the proper structure.\nProblematic input: {error['input']}\nFix: {error['msg']}"
            )
        os.environ["MOCKAI_RESPONSES"] = responses.name

    os.environ["MOCKAI_EMBEDDING_SIZE"] = str(embedding_size)

    print("Starting MockAI server ...")  # noqa: T201
    subprocess.run(
        [
            "uvicorn",
            "--app-dir",
            f"{dir_path}",
            "main:app",
            "--host",
            host,
            "--port",
            str(port),
            "--log-config",
            f"{dir_path}/logging_conf.yaml",
        ]
    )



import json
import logging
from uuid import uuid4

from pydantic import ValidationError

from mockai.dependencies import ResponseFile
from mockai.models.common import FunctionOutput

from .models import anthropic_tool

_logger = logging.getLogger(__name__)


def json_response(response_array, model: str):
    response = {
        "id": f"msg_{uuid4().hex}",
        "type": "message",
        "model": model,
        "role": "assistant",
        "content": response_array,
        "stop_reason": "end_turn",
        "stop_sequence": None,
        "usage": {
            "input_tokens": 0,
            "output_tokens": 0,
        },
    }
    return response


def stream_chunk(event: str, data: dict):
    return f"event: {event}\ndata: {json.dumps(data)}\n\n"


def streaming_response(response_array, model: str):
    id = f"msg_{uuid4().hex}"

    message_start = {
        "type": "message_start",
        "message": {
            "id": id,
            "type": "message",
            "role": "assistant",
            "content": [],
            "model": model,
            "stop_reason": None,
            "stop_sequence": None,
            "usage": {"input_tokens": 0, "output_tokens": 0},
        },
    }
    yield stream_chunk("message_start", message_start)

    yield stream_chunk("ping", {"type": "ping"})

    is_function = False

    for idx, event in enumerate(response_array):
        try:
            is_function = True if event["type"] == "tool_use" else False
        except Exception:
            _logger.exception("unexpected error: durring streaming response")

        block_start = {
            "type": "content_block_start",
            "index": idx,
            "content_block": {
                "type": event["type"],
                "id": event["id"],
                "name": event["name"],
                "input": {},
            }
            if is_function
            else {"type": "text", "text": ""},
        }
        yield stream_chunk("content_block_start", block_start)

        if is_function:
            iterator = json.dumps(event["input"])
        else:
            iterator = event["text"]

        for char in iterator:
            chunk = {"type": "content_block_delta", "index": idx}

            if is_function:
                chunk["delta"] = {"type": "input_json_delta", "partial_json": char}
            else:
                chunk["delta"] = {"type": "text_delta", "text": char}

            yield stream_chunk("content_block_delta", chunk)

        block_stop = {"type": "content_block_stop", "index": idx}
        yield stream_chunk("content_block_stop", block_stop)

    message_delta = {
        "type": "message_delta",
        "delta": {"stop_reason": "tool_use" if is_function else "end_turn"},
        "usage": {"output_tokens": 0},
    }
    yield stream_chunk("message_delta", message_delta)

    yield stream_chunk("message_stop", {"type": "message_stop"})


async def generate_anthropic_response(
    payload: dict,
    file: ResponseFile,
    mock_response: str | None,
):
    model = payload["model"]
    stream = payload.get("stream")
    # Get content from last message
    content = None
    for message in payload["messages"][::-1]:
        if message["role"] == "user":
            content = message["content"]
            break

    if content is None:
        content = payload["messages"][-1]["content"]

    # If content is list of objects, get text from first object where type == text
    if type(content) is list:
        for obj in content:
            if obj["type"] == "text":
                content = obj["text"]
                break
        else:
            raise ValueError(
                "Content array must include at least one object with 'type' = 'text'",
            )

    # Set initial response_array
    response_array = [{"type": "text", "text": content}]

    # Check predetermined responses for matching inputs
    found_predetermined = False
    if file is not None:
        response = file.find_matching_or_none(origin="anthropic", payload=payload)
        if response is not None:
            if response.type == "text":
                response_array = [{"type": "text", "text": response.output}]
                found_predetermined = True

            elif response.type == "function":
                if isinstance(output := response.output, str):
                    raise ValueError("Impossible state")

                response_array = [anthropic_tool(m) for m in output._to_list()]
                found_predetermined = True
    _logger.info(
        "Predetermined response %s found", "not" if not found_predetermined else ""
    )

    # Check if a mock response was passed in header
    if mock_response is not None:
        if found_predetermined:
            _logger.info(
                "Overriding predetermined response with mock response from header"
            )
        else:
            _logger.info("Using mock response from header")
        try:
            if mock_response[:2] == "f:":
                function_output = FunctionOutput.model_validate_json(mock_response[2:])
                response_array = [anthropic_tool(function_output)]
            else:
                response_array = [{"type": "text", "text": mock_response}]

        except (ValidationError, json.JSONDecodeError) as e:
            response_array = [{"type": "text", "text": str(e)}]

    if stream is None or stream is False:
        return json_response(response_array, model)
    else:
        return streaming_response(response_array, model)



import json

import pytest
from anthropic import AsyncStream, Stream
from anthropic.types import Message, TextBlock, ToolUseBlock
from httpx import AsyncClient as HTTPXACLIENT
from httpx import Client as HTTPXClient

from mockai.anthropic import Anthropic, AsyncAnthropic, AsyncClient, Client

CustomSyncClient = Anthropic(http_client=HTTPXClient(http2=True))
CustomAsyncClient = AsyncAnthropic(http_client=HTTPXACLIENT(http2=True))


# Fixtures
@pytest.fixture(params=[Anthropic(), Client(), CustomSyncClient])
def client_x(request):
    return request.param


@pytest.fixture(params=[AsyncAnthropic(), AsyncClient(), CustomAsyncClient])
def async_client_x(request):
    return request.param


# Tests
def test_message(client_x):
    completion = client_x.messages.create(
        model="mock", messages=[{"role": "user", "content": "Hello!"}], max_tokens=1024
    )
    content = completion.content[0]
    assert isinstance(content, TextBlock)
    assert content.text == "Hello!"


def test_predefined_message(client_x):
    completion = client_x.messages.create(
        model="mock",
        messages=[{"role": "user", "content": "How are ya?"}],
        max_tokens=1024,
    )
    content = completion.content[0]
    assert isinstance(content, TextBlock)
    assert content.text == "I'm fine, thank u 😊. How about you?"


def test_message_stream(client_x):
    generator = client_x.messages.create(
        model="mock",
        messages=[{"role": "user", "content": "How are ya?"}],
        max_tokens=1024,
        stream=True,
    )
    assert isinstance(generator, Stream)
    buffer = ""
    for chunk in generator:
        if chunk.type == "content_block_delta":
            buffer += chunk.delta.text
    assert buffer == "I'm fine, thank u 😊. How about you?"


def test_tool_call(client_x):
    response = client_x.messages.create(
        model="mock",
        messages=[{"role": "user", "content": "Where's my order?"}],
        max_tokens=1024,
    )
    tool = response.content[0]
    assert isinstance(tool, ToolUseBlock)
    assert tool.name == "get_delivery_date"
    assert tool.input == {"order_id": "1337", "order_loc": ["New York", "Mexico City"]}


def test_tool_call_stream(client_x):
    generator = client_x.messages.create(
        model="mock",
        messages=[{"role": "user", "content": "Where's my order?"}],
        max_tokens=1024,
        stream=True,
    )
    assert isinstance(generator, Stream)
    buffer = ""
    for chunk in generator:
        if chunk.type == "content_block_delta":
            buffer += chunk.delta.partial_json
    assert json.loads(buffer) == {
        "order_id": "1337",
        "order_loc": ["New York", "Mexico City"],
    }


@pytest.mark.asyncio
async def test_async_message(async_client_x):
    completion = await async_client_x.messages.create(
        model="mock", messages=[{"role": "user", "content": "Hello!"}], max_tokens=1024
    )
    assert isinstance(completion, Message)
    content = completion.content[0]
    assert isinstance(content, TextBlock)
    assert content.text == "Hello!"


@pytest.mark.asyncio
async def test_predefined_async_message(async_client_x):
    completion = await async_client_x.messages.create(
        model="mock",
        messages=[{"role": "user", "content": "Where's my order?"}],
        max_tokens=1024,
    )
    assert isinstance(completion, Message)
    assert isinstance(completion.content[0], ToolUseBlock)
    assert completion.content[0].name == "get_delivery_date"
    assert completion.content[0].input == {
        "order_id": "1337",
        "order_loc": ["New York", "Mexico City"],
    }


@pytest.mark.asyncio
async def test_async_message_stream(async_client_x):
    generator = await async_client_x.messages.create(
        model="mock",
        messages=[{"role": "user", "content": "How are ya?"}],
        max_tokens=1024,
        stream=True,
    )
    assert isinstance(generator, AsyncStream)
    buffer = ""
    async for chunk in generator:
        if chunk.type == "content_block_delta":
            buffer += chunk.delta.text
    assert buffer == "I'm fine, thank u 😊. How about you?"


@pytest.mark.asyncio
async def test_async_tool_call(async_client_x):
    response = await async_client_x.messages.create(
        model="mock",
        messages=[{"role": "user", "content": "Where's my order?"}],
        max_tokens=1024,
    )
    tool = response.content[0]
    assert isinstance(tool, ToolUseBlock)
    assert tool.name == "get_delivery_date"
    assert tool.input == {"order_id": "1337", "order_loc": ["New York", "Mexico City"]}


@pytest.mark.asyncio
async def test_async_tool_call_stream(async_client_x):
    generator = await async_client_x.messages.create(
        model="mock",
        messages=[{"role": "user", "content": "Where's my order?"}],
        max_tokens=1024,
        stream=True,
    )
    assert isinstance(generator, AsyncStream)
    buffer = ""
    async for chunk in generator:
        if chunk.type == "content_block_delta":
            buffer += chunk.delta.partial_json
    assert json.loads(buffer) == {
        "order_id": "1337",
        "order_loc": ["New York", "Mexico City"],
    }


@pytest.mark.asyncio
async def test_user_message_is_not_last_no_match(async_client_x):
    completion = await async_client_x.messages.create(
        model="mock",
        messages=[
            {"role": "user", "content": "Where's my json you do not know?"},
            {"role": "assistant", "content": "your json is here:"},
        ],
        max_tokens=1024,
    )
    content = completion.content[0]
    assert isinstance(content, TextBlock)
    assert content.text == "Where's my json you do not know?"


@pytest.mark.asyncio
async def test_user_message_is_not_last_and_is_matched():
    async_client = AsyncAnthropic()
    completion = await async_client.messages.create(
        model="mock",
        messages=[
            {"role": "user", "content": "Where's my json you know?"},
            {"role": "assistant", "content": "your json is here:"},
        ],
        max_tokens=1024,
    )
    content = completion.content[0]
    assert isinstance(content, TextBlock)
    assert content.text == "{'fake': 'json'}"


@pytest.mark.asyncio
async def test_cached_system_prompt_allowed(async_client_x):
    # https://docs.anthropic.com/en/docs/build-with-claude/prompt-caching
    response = await async_client_x.messages.create(
        model="mock",
        system=[
            {
                "type": "text",
                "text": "test_system_prompt",
                "cache_control": {
                    "type": "ephemeral"
                },  # Cache the static system prompt
            }
        ],
        messages=[{"role": "user", "content": "Where's my orda?"}],
        max_tokens=1024,
    )
    content = response.content[0]
    assert isinstance(content, TextBlock)
    assert content.text == "ORDER MATCHED BASED ON SYSTEM PROMPT"



import pytest
from openai.types import CreateEmbeddingResponse
from openai.types.chat import (
    ChatCompletion,
    ChatCompletionChunk,
    ChatCompletionMessage,
    ChatCompletionMessageToolCall,
)
from openai.types.chat.chat_completion_chunk import ChoiceDelta, ChoiceDeltaToolCall

from mockai.openai import (
    AsyncAzureOpenAI,
    AsyncClient,
    AsyncOpenAI,
    AzureOpenAI,
    Client,
    OpenAI,
)


@pytest.mark.parametrize("client", [OpenAI(), Client(), AzureOpenAI()])
def test_openai_chat_completion(client):
    completion = client.chat.completions.create(
        model="mock", messages=[{"role": "user", "content": "Hello!"}]
    )
    assert isinstance(completion, ChatCompletion)
    assert isinstance(completion.choices[0].message, ChatCompletionMessage)
    assert isinstance(completion.choices[0].message.content, str)


@pytest.mark.parametrize("client", [OpenAI(), Client(), AzureOpenAI()])
def test_openai_chat_programmed_completion(client):
    completion = client.chat.completions.create(
        model="mock", messages=[{"role": "user", "content": "How are ya?"}]
    )
    assert isinstance(completion, ChatCompletion)
    assert isinstance(completion.choices[0].message, ChatCompletionMessage)
    assert (
        completion.choices[0].message.content == "I'm fine, thank u 😊. How about you?"
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("client", [AsyncOpenAI(), AsyncClient(), AsyncAzureOpenAI()])
async def test_async_openai_chat_completion(client):
    completion = await client.chat.completions.create(
        model="mock", messages=[{"role": "user", "content": "Hello!"}]
    )
    assert isinstance(completion, ChatCompletion)
    assert isinstance(completion.choices[0].message, ChatCompletionMessage)
    assert isinstance(completion.choices[0].message.content, str)


@pytest.mark.asyncio
@pytest.mark.parametrize("client", [AsyncOpenAI(), AsyncClient(), AsyncAzureOpenAI()])
async def test_async_openai_chat_programmed_completion(client):
    completion = await client.chat.completions.create(
        model="mock",
        messages=[{"role": "user", "content": "Where's my order?"}],
    )
    assert isinstance(completion, ChatCompletion)
    assert isinstance(completion.choices[0].message, ChatCompletionMessage)
    assert (
        completion.choices[0].message.tool_calls[0].function.name == "get_delivery_date"  # type: ignore
    )
    assert (
        completion.choices[0].message.tool_calls[0].function.arguments  # type: ignore
        == {"order_id": "1337", "order_loc": ["New York", "Mexico City"]}
    )


@pytest.mark.parametrize("client", [OpenAI(), Client(), AzureOpenAI()])
def test_openai_chat_completion_stream(client):
    response = client.chat.completions.create(
        model="mock", messages=[{"role": "user", "content": "Hello!"}], stream=True
    )
    completion = next(response)
    assert isinstance(completion, ChatCompletionChunk)
    assert isinstance(completion.choices[0].delta.content, str)


@pytest.mark.asyncio
@pytest.mark.parametrize("client", [AsyncOpenAI(), AsyncClient(), AsyncAzureOpenAI()])
async def test_async_openai_chat_completion_stream(client):
    response = await client.chat.completions.create(
        model="mock", messages=[{"role": "user", "content": "Hello!"}], stream=True
    )
    completion = await anext(response)
    assert isinstance(completion, ChatCompletionChunk)
    assert isinstance(completion.choices[0].delta.content, str)


@pytest.mark.parametrize("client", [OpenAI(), Client(), AzureOpenAI()])
def test_openai_function_call(client):
    completion = client.chat.completions.create(
        model="mock", messages=[{"role": "user", "content": "Where's my order?"}]
    )
    assert isinstance(completion, ChatCompletion)
    assert isinstance(completion.choices[0].message, ChatCompletionMessage)
    assert isinstance(
        completion.choices[0].message.tool_calls[0],  # type: ignore
        ChatCompletionMessageToolCall,
    )


@pytest.mark.parametrize("client", [OpenAI(), Client(), AzureOpenAI()])
def test_openai_function_call_followup(client):
    completion = client.chat.completions.create(
        model="mock",
        messages=[
            {"role": "user", "content": "Where's my order?"},
            {
                "role": "assistant",
                "tool_calls": [
                    {
                        "id": "call_123",
                        "type": "function",
                        "function": {
                            "name": "get_delivery_date",
                            "arguments": "{'order_id': '1337'}",
                        },
                    }
                ],
            },
            {
                "role": "tool",
                "content": "{'order_id': '1337', 'delivery_date': '25/10/2024'}",
                "tool_call_id": "call_123",
            },
        ],
    )
    assert isinstance(completion, ChatCompletion)

    message = completion.choices[0].message
    assert isinstance(message, ChatCompletionMessage)
    assert message.content == "Your order will arrive the 25th of October."


@pytest.mark.parametrize("client", [OpenAI(), Client(), AzureOpenAI()])
def test_openai_programmed_function_call(client):
    completion = client.chat.completions.create(
        model="mock", messages=[{"role": "user", "content": "Where's my order?"}]
    )
    assert isinstance(completion, ChatCompletion)
    assert isinstance(completion.choices[0].message, ChatCompletionMessage)
    assert isinstance(
        completion.choices[0].message.tool_calls[0],  # type: ignore
        ChatCompletionMessageToolCall,
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("client", [AsyncOpenAI(), AsyncClient(), AsyncAzureOpenAI()])
async def test_async_openai_function_call(client):
    completion = await client.chat.completions.create(
        model="mock", messages=[{"role": "user", "content": "Where's my order?"}]
    )
    assert isinstance(completion, ChatCompletion)
    assert isinstance(completion.choices[0].message, ChatCompletionMessage)
    assert isinstance(
        completion.choices[0].message.tool_calls[0],  # type: ignore
        ChatCompletionMessageToolCall,
    )


@pytest.mark.parametrize("client", [OpenAI(), Client(), AzureOpenAI()])
def test_openai_function_call_stream(client):
    response = client.chat.completions.create(
        model="mock",
        messages=[{"role": "user", "content": "Where's my order?"}],
        stream=True,
    )
    completion = next(response)
    assert isinstance(completion, ChatCompletionChunk)
    assert isinstance(completion.choices[0].delta, ChoiceDelta)
    assert isinstance(
        completion.choices[0].delta.tool_calls[0],  # type: ignore
        ChoiceDeltaToolCall,
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("client", [AsyncOpenAI(), AsyncClient(), AsyncAzureOpenAI()])
async def test_async_openai_function_call_stream(client):
    response = await client.chat.completions.create(
        model="mock",
        messages=[{"role": "user", "content": "Where's my order?"}],
        stream=True,
    )
    completion = await anext(response)
    assert isinstance(completion, ChatCompletionChunk)
    assert isinstance(completion.choices[0].delta, ChoiceDelta)
    assert isinstance(
        completion.choices[0].delta.tool_calls[0],  # type: ignore
        ChoiceDeltaToolCall,
    )


# Embeddings


@pytest.mark.parametrize("client", [OpenAI(), Client(), AzureOpenAI()])
def test_openai_embedding(client):
    response = client.embeddings.create(model="mock", input="hello")
    assert isinstance(response, CreateEmbeddingResponse)
    assert len(response.data) == 1
    assert len(response.data[0].embedding) == 1536


@pytest.mark.asyncio
@pytest.mark.parametrize("client", [AsyncOpenAI(), AsyncClient(), AsyncAzureOpenAI()])
async def test_async_embedding(client):
    response = await client.embeddings.create(model="mock", input="hello")
    assert isinstance(response, CreateEmbeddingResponse)
    assert len(response.data) == 1
    assert len(response.data[0].embedding) == 1536


@pytest.mark.parametrize("client", [OpenAI(), Client(), AzureOpenAI()])
def test_openai_embedding_list(client):
    input_list = ["hiii", "my", "name", "is", "joe"]

    response = client.embeddings.create(model="mock", input=input_list)
    assert isinstance(response, CreateEmbeddingResponse)
    assert len(response.data) == 5

    for i, data in enumerate(response.data):
        assert data.index == i
        assert len(data.embedding) == 1536


@pytest.mark.asyncio
@pytest.mark.parametrize("client", [AsyncOpenAI(), AsyncClient(), AsyncAzureOpenAI()])
async def test_async_embedding_list(client):
    input_list = ["hiii", "my", "name", "is", "joe"]

    response = await client.embeddings.create(model="mock", input=input_list)
    assert isinstance(response, CreateEmbeddingResponse)
    assert len(response.data) == 5

    for i, data in enumerate(response.data):
        assert data.index == i
        assert len(data.embedding) == 1536


@pytest.mark.asyncio
@pytest.mark.parametrize("client", [AsyncOpenAI(), AsyncClient(), AsyncAzureOpenAI()])
async def test_async_openai_user_message_is_not_last_not_matched(client):
    completion = await client.chat.completions.create(
        model="mock",
        messages=[
            {"role": "user", "content": "Where's my json you do not know?"},
            {"role": "assistant", "content": "your json is here:"},
        ],
    )

    assert isinstance(completion, ChatCompletion)

    message = completion.choices[0].message
    assert isinstance(message, ChatCompletionMessage)
    assert message.content == "Where's my json you do not know?"


@pytest.mark.asyncio
@pytest.mark.parametrize("client", [AsyncOpenAI(), AsyncClient(), AsyncAzureOpenAI()])
async def test_async_openai_user_message_is_not_last_matched(client):
    completion = await client.chat.completions.create(
        model="mock",
        messages=[
            {"role": "user", "content": "Where's my json you know?"},
            {"role": "assistant", "content": "your json is here:"},
        ],
    )

    assert isinstance(completion, ChatCompletion)

    message = completion.choices[0].message
    assert isinstance(message, ChatCompletionMessage)
    assert message.content == "{'fake': 'json'}"



import warnings
from functools import partial

from mockai.constants import API_KEY, BASE_ENDPOINT, NOT_AVAILABLE

from .router import anthropic_router as _router
from .services import generate_anthropic_response as _generate_anthropic_response

__all__ = [
    "Anthropic",
    "AsyncAnthropic",
    "Client",
    "AsyncClient",
    "_router",
    "_generate_anthropic_response",
]

ANTHROPIC_ENDPOINT = BASE_ENDPOINT + "/anthropic"

Anthropic = NOT_AVAILABLE
AsyncAnthropic = NOT_AVAILABLE

Client = NOT_AVAILABLE
AsyncClient = NOT_AVAILABLE

try:
    from anthropic import (
        Anthropic,
        AsyncAnthropic,
        AsyncClient,
        Client,
    )

    Anthropic = partial(Anthropic, base_url=ANTHROPIC_ENDPOINT, api_key=API_KEY)
    AsyncAnthropic = partial(
        AsyncAnthropic, base_url=ANTHROPIC_ENDPOINT, api_key=API_KEY
    )

    Client = partial(Client, base_url=ANTHROPIC_ENDPOINT, api_key=API_KEY)
    AsyncClient = partial(AsyncClient, base_url=ANTHROPIC_ENDPOINT, api_key=API_KEY)
except (ImportError, TypeError):
    warnings.warn("Anthropic SDK not installed, anthropic clients are not available")



from __future__ import annotations

from typing import Literal
from uuid import uuid4

from pydantic import BaseModel, ConfigDict, model_validator
from typing_extensions import TypedDict

from mockai.models.common import FunctionOutput

roles = Literal["user", "assistant"]

content_types = Literal["text", "image", "tool_result", "tool_use"]


class CacheControl(TypedDict, total=False):
    type: Literal["ephemeral"]


class CachedSystemPrompt(BaseModel):
    type: Literal["text"]
    text: str
    cache_control: CacheControl | None


class Content(BaseModel):
    type: content_types
    text: str | None = None
    source: dict | None = None
    tool_use_id: str | None = None
    content: str | None = None

    @model_validator(mode="after")
    def check_fields(self) -> Content:
        match self.type:
            case "text":
                if self.text is None:
                    raise ValueError("text field required")
            case "image":
                if self.source is None:
                    raise ValueError("image field required")
            case "tool_result":
                if self.tool_use_id is None or self.content is None:
                    raise ValueError(
                        "tool_use_id and content field required for content of type 'tool_result'"
                    )
        return self


class Message(BaseModel):
    role: roles
    content: str | list[Content]


class Payload(BaseModel):
    model: str
    max_tokens: int
    messages: list[Message]
    stream: bool | None = None
    system: str | list[CachedSystemPrompt] | None = None

    model_config = ConfigDict(extra="allow")


class AnthropicTool(TypedDict):
    id: str
    type: str
    input: dict
    name: str


def anthropic_tool(function: FunctionOutput):
    return AnthropicTool(
        {
            "id": f"toolu_{uuid4().hex}",
            "type": "tool_use",
            "input": function.arguments,
            "name": function.name,
        }
    )



import os

import aiofiles
from fastapi import APIRouter
from fastapi.exceptions import HTTPException
from fastapi.staticfiles import StaticFiles
from pydantic import BaseModel

from mockai.dependencies import ResponseFile
from mockai.models.json_file.models import PreDeterminedResponse

gui_router = APIRouter(prefix="/api/responses")

gui_router.mount(
    "/gui",
    StaticFiles(
        directory=f"{os.path.dirname(os.path.realpath(__file__))}/static", html=True
    ),
    name="gui",
)


@gui_router.get("/create")
async def create_response(responses: ResponseFile):
    if responses is None:
        raise HTTPException(400, "No response file to update.")

    responses.append(PreDeterminedResponse(type="text", input="", output=""))

    if file := os.getenv("MOCKAI_RESPONSES"):
        async with aiofiles.open(file, "w") as f:
            await f.write(responses.model_dump_json(indent=2))
        return responses
    else:
        raise ValueError("No response file set.")


@gui_router.get("/read")
async def get_responses(file: ResponseFile):
    return file


class ResponseUpdate(BaseModel):
    number: int
    new_response: PreDeterminedResponse


@gui_router.post("/update")
async def update_response(data: ResponseUpdate, responses: ResponseFile):
    if responses is None:
        raise HTTPException(400, "No response file to update.")

    responses[data.number - 1] = data.new_response

    if file := os.getenv("MOCKAI_RESPONSES"):
        async with aiofiles.open(file, "w") as f:
            await f.write(responses.model_dump_json(indent=2))
        return responses
    else:
        raise ValueError("No response file set.")


@gui_router.delete("/delete")
async def delete_response(number: int, responses: ResponseFile):
    if responses is None:
        raise HTTPException(400, "No response file to update.")

    responses.pop(number - 1)

    if file := os.getenv("MOCKAI_RESPONSES"):
        async with aiofiles.open(file, "w") as f:
            await f.write(responses.model_dump_json(indent=2))
        return responses
    else:
        raise ValueError("No response file set.")



from __future__ import annotations

from typing import Annotated, Literal, TypeAlias

from pydantic import BaseModel, Field, model_validator

from mockai.models.common import FunctionOutput, FunctionOutputs

ResponseType: TypeAlias = Literal["text", "function"]


class InputMatcher(BaseModel):
    system_prompt_name: Annotated[str | None, Field(default=None)]
    role: Annotated[str | None, Field(default=None)]
    content: str
    offset: Annotated[int, Field(default=-1)]

    def is_matching_payload(
        self, origin: str, payload: dict, system_prompts: dict[str, str]
    ) -> bool:
        msg_count = len(messages := payload["messages"])
        if not -msg_count <= self.offset < msg_count:
            return False
        message = messages[self.offset]
        if origin == "anthropic":
            if isinstance(message["content"], list):
                found = False
                for content in message["content"]:
                    if self.content == content["text"]:
                        found = True
                        break
                if not found:
                    return False
            else:
                if self.content != message["content"]:
                    return False

            if self.role is not None and self.role != message["role"]:
                return False
            if self.system_prompt_name is not None:
                if isinstance(system := payload["system"], list):
                    prompt = system_prompts[self.system_prompt_name]
                    return any(sys_prompt["text"] == prompt for sys_prompt in system)
                return system == system_prompts[self.system_prompt_name]
            return True
        else:
            if isinstance(message["content"], list):
                found = False
                for content in message["content"]:
                    if self.content == content["text"]:
                        found = True
                        break
                if not found:
                    return False
            else:
                if self.content != message["content"]:
                    return False
            if self.role is not None and self.role != message["role"]:
                return False
            return True


class PreDeterminedResponse(BaseModel):
    type: ResponseType
    input: InputMatcher | str
    output: str | FunctionOutput | FunctionOutputs

    @model_validator(mode="after")
    def verify_structure(self) -> PreDeterminedResponse:
        if self.type == "function":
            try:
                if isinstance(self.output, list):
                    checks = [isinstance(f, FunctionOutput) for f in self.output]
                    assert all(checks)
                else:
                    assert isinstance(self.output, FunctionOutput)
            except AssertionError:
                raise ValueError(
                    "When a response is of type 'function', the output must be a single FunctionOutput object or an array of FunctionOutput objects"
                )
        elif self.type == "text":
            if isinstance(self.output, str) is False:
                raise ValueError(
                    "When a response is of type 'text', the output must be a string."
                )
        return self

    def response_matches(
        self, origin: str, payload: dict, system_prompts: dict[str, str]
    ) -> bool:
        if isinstance(self.input, str):
            return self.input == payload["messages"][-1]["content"]
        else:
            return self.input.is_matching_payload(origin, payload, system_prompts)


class PreDeterminedResponses(BaseModel):
    responses: Annotated[list[PreDeterminedResponse], Field(default=[])]
    system_prompts: Annotated[dict[str, str], Field(default={})]

    @model_validator(mode="after")
    def _verify_responses(self) -> PreDeterminedResponses:
        glob_errors = []
        temp_errors = [
            response.input.system_prompt_name
            for response in self.responses
            if (
                isinstance(response.input, InputMatcher)
                and response.input.system_prompt_name is not None
                and response.input.system_prompt_name not in self.system_prompts
            )
        ]
        if len(temp_errors) > 0:
            glob_errors.append(
                f"Following system prompt missing for the following system_prompt_names: {temp_errors}"
            )
            temp_errors = []

        if len(glob_errors) > 0:
            raise ValueError(
                "found following errors in file:\n" + "\n".join(glob_errors)
            )
        return self

    def find_matching_or_none(
        self, origin: str, payload: dict
    ) -> PreDeterminedResponse | None:
        for response in self.responses:
            if response.response_matches(origin, payload, self.system_prompts):
                return response
        return None

    # todo add chaching that will be flushed when file changes



from typing import Any, Literal, TypeAlias

from pydantic import BaseModel, RootModel

ResponseType: TypeAlias = Literal["text", "function"]


class FunctionOutput(BaseModel):
    name: str
    arguments: dict[str, Any]

    def _to_dict_list(self):
        return [self.model_dump()]

    def _to_list(self):
        return [self]


class FunctionOutputs(RootModel):
    root: list[FunctionOutput]

    def __iter__(self):  # type: ignore
        return iter(self.root)

    def _to_dict_list(self):
        return self.model_dump()

    def _to_list(self):
        return self.root



import json
import logging
import random
from itertools import zip_longest
from time import time
from typing import cast
from uuid import uuid4

from pydantic import ValidationError

from mockai.dependencies import ResponseFile
from mockai.models.json_file.models import PreDeterminedResponse

_log = logging.getLogger(__name__)


def json_response(content: str | None, model: str, tool_calls: list[dict] | None):
    response = {
        "id": f"chatcmpl-{uuid4().hex}",
        "object": "chat.completion",
        "created": int(time()),
        "model": model,
        "system_fingerprint": "mock",
        "choices": [
            {
                "index": 0,
                "message": {
                    "role": "assistant",
                    "content": content,
                    "tool_calls": tool_calls,
                },
                "logprobs": None,
                "finish_reason": "stop",
            }
        ],
        "usage": {
            "prompt_tokens": 0,
            "completion_tokens": 0,
            "total_tokens": 0,
            "completion_tokens_details": {"reasoning_tokens": 0},
        },
    }
    return response


def streaming_response(content: str | None, model: str, tool_calls: list[dict] | None):
    id = f"chatcmpl-{uuid4().hex}"

    if content is not None:
        iterator = content
    elif tool_calls is not None:
        iterator = zip_longest(
            *[
                list(json.dumps(tool_call["function"]["arguments"]))
                for tool_call in tool_calls
            ]
        )
    else:
        raise ValueError("Either content or tool_calls must not be None")

    for i in iterator:
        chunk = {
            "id": id,
            "object": "chat.completion.chunk",
            "created": int(time()),
            "model": model,
            "system_fingerprint": "mock",
            "choices": [
                {
                    "index": 0,
                    "delta": {
                        "role": "assistant",
                        "content": i if content is not None else None,
                        "tool_calls": [
                            {
                                "id": tool_call["id"],
                                "type": tool_call["type"],
                                "function": {
                                    "name": tool_call["function"]["name"],
                                    "arguments": i[n],
                                },
                            }
                            for n, tool_call in enumerate(tool_calls)
                        ]
                        if tool_calls is not None
                        else None,
                    },
                    "logprobs": None,
                    "finish_reason": None,
                }
            ],
        }
        yield f"data: {json.dumps(chunk)}\n\n"

    yield "data: [DONE]\n\n"


def response_struct_to_openai_format(response: PreDeterminedResponse):
    if response.type == "text":
        content = response.output
        tool_calls = None
    elif response.type == "function":
        content = None

        if isinstance(response.output, str):
            raise ValueError("Impossible state")

        tool_calls = response.output._to_dict_list()

        for tool_call in tool_calls:
            tool_call["id"] = str(uuid4())
            tool_call["type"] = "function"
            function = {
                "name": tool_call.pop("name"),
                "arguments": tool_call.pop("arguments"),
            }
            tool_call["function"] = function
    else:
        raise ValueError("unreachable")

    return content, tool_calls


async def generate_openai_completion_response(
    payload: dict,
    responses: ResponseFile,
    mock_response: str | None,
):
    model = payload["model"]
    stream = payload.get("stream")
    content = None
    for message in payload["messages"][::-1]:
        if message["role"] == "user":
            content = message["content"]
            break
    if content is None:
        content = payload["messages"][-1]["content"]

    if content is None:
        raise ValueError("Content from last message cannot be None")
    tool_calls = None

    if isinstance(content, list):
        for obj in content:
            if obj["type"] == "text":
                content = obj["text"]
                break
        else:
            raise ValueError(
                "Content array must include at least one object with 'type' = 'text'",
            )
    found_predetermined_response = False
    if responses is not None:
        response = responses.find_matching_or_none(origin="openai", payload=payload)
        if response is not None:
            content, tool_calls = response_struct_to_openai_format(response)
            found_predetermined_response = True

    _log.info(
        "predetermined response %s found",
        "not" if not found_predetermined_response else "",
    )

    if mock_response is not None:
        if found_predetermined_response:
            _log.info(
                "Overriding predetermined response with mock response from header"
            )
        else:
            _log.info("Using mock response from header")
        try:
            is_function = mock_response[:2] == "f:"

            r_type = "function" if is_function else "text"

            if is_function:
                output = json.loads(mock_response[2:])
            else:
                output = mock_response

            header_mock_response = PreDeterminedResponse(
                type=r_type, input="None", output=output
            )
            content, tool_calls = response_struct_to_openai_format(header_mock_response)
        except (ValidationError, json.JSONDecodeError) as e:
            content = str(e)

    content = cast(str, content)

    if stream is None or stream is False:
        return json_response(content, model, tool_calls)
    else:
        return streaming_response(content, model, tool_calls)


async def generate_openai_embeddings_response(embedding_size: int, payload: dict):
    if isinstance(input := payload["input"], str):
        input_list = [input]
    else:
        input_list = input

    embedding_range = range(embedding_size)
    input_range = range(len(input_list))

    return {
        "object": "list",
        "data": [
            {
                "object": "embedding",
                "embedding": [random.uniform(-1, 1) for _ in embedding_range],
                "index": number,
            }
            for number in input_range
        ],
        "model": payload["model"],
        "usage": {
            "prompt_tokens": 0,
            "total_tokens": 0,
        },
    }



from __future__ import annotations

from typing import Literal

from pydantic import BaseModel, ConfigDict, model_validator


class Content(BaseModel):
    type: Literal["text", "image_url"]
    text: str | None = None
    image_url: dict | None = None

    @model_validator(mode="after")
    def check_fields(self) -> Content:
        if self.type == "text":
            if self.text is None:
                raise ValueError("text field required")
        elif self.type == "image_url":
            if self.image_url is None:
                raise ValueError("image_url field required")
            url = self.image_url.get("url", None)
            if isinstance(url, str):
                raise ValueError('image_url dict must contain "url" key of type string')
        return self


class Message(BaseModel):
    role: Literal["system", "user", "assistant", "tool"]
    content: str | list[Content] | None = None

    model_config = ConfigDict(extra="allow")


class Payload(BaseModel):
    model: str
    messages: list[Message]
    stream: bool | None = None

    model_config = ConfigDict(extra="allow")

    @model_validator(mode="after")
    def check_messages(self) -> Payload:
        if len(self.messages) == 0:
            raise ValueError("messages array can't be empty.")
        return self


class EmbeddingPayload(BaseModel):
    model: str
    input: str | list[str]

    model_config = ConfigDict(extra="allow")



import warnings
from functools import partial

from mockai.constants import API_KEY, BASE_ENDPOINT, NOT_AVAILABLE

from .router import openai_router as _router
from .services import (
    generate_openai_completion_response as _generate_openai_completion_response,
)
from .services import (
    generate_openai_embeddings_response as _generate_openai_embeddings_response,
)

__all__ = [
    "OpenAI",
    "AsyncOpenAI",
    "Client",
    "AsyncClient",
    "AzureOpenAI",
    "AzureOpenAI",
    "_router",
    "_generate_openai_completion_response",
    "_generate_openai_embeddings_response",
]

OPENAI_ENDPOINT = BASE_ENDPOINT + "/openai"

OpenAI = NOT_AVAILABLE
AsyncOpenAI = NOT_AVAILABLE

Client = NOT_AVAILABLE
AsyncClient = NOT_AVAILABLE

AzureOpenAI = NOT_AVAILABLE
AsyncAzureOpenAI = NOT_AVAILABLE

try:
    from openai import (
        AsyncAzureOpenAI,
        AsyncClient,
        AsyncOpenAI,
        AzureOpenAI,
        Client,
        OpenAI,
    )

    OpenAI = partial(OpenAI, base_url=OPENAI_ENDPOINT, api_key=API_KEY)
    AsyncOpenAI = partial(AsyncOpenAI, base_url=OPENAI_ENDPOINT, api_key=API_KEY)

    Client = partial(Client, base_url=OPENAI_ENDPOINT, api_key=API_KEY)
    AsyncClient = partial(AsyncClient, base_url=OPENAI_ENDPOINT, api_key=API_KEY)

    AzureOpenAI = partial(
        AzureOpenAI, base_url=OPENAI_ENDPOINT, api_key=API_KEY, api_version="None"
    )
    AsyncAzureOpenAI = partial(
        AsyncAzureOpenAI, base_url=OPENAI_ENDPOINT, api_key=API_KEY, api_version="None"
    )
except (ImportError, TypeError):
    warnings.warn("OpenAI SDK not installed, openai clients are not available")
